from graphql.pyutils import is_collection

from .sync_future import SyncFuture


class SyncDataLoader:
    def __init__(self, batch_load_fn):
        self._batch_load_fn = batch_load_fn
        self._cache = {}
        self._queue = []

    def load(self, key):
        try:
            return self._cache[key]
        except KeyError:
            future = SyncFuture()
            needs_dispatch = not self._queue
            self._queue.append((key, future))
            if needs_dispatch:
                future.deferred_callback = self.dispatch_queue
            self._cache[key] = future
            return future

    def clear(self, key):
        self._cache.pop(key, None)

    def dispatch_queue(self):
        queue = self._queue
        if not queue:
            return
        self._queue = []

        keys = [item[0] for item in queue]
        values = self._batch_load_fn(keys)
        if not is_collection(values) or len(keys) != len(values):
            raise ValueError("The batch loader does not return an expected result")

        try:
            for (key, future), value in zip(queue, values):
                if isinstance(value, Exception):
                    future.set_exception(value)
                else:
                    future.set_result(value)
        except Exception as error:
            for key, future in queue:
                self.clear(key)
                if not future.done():
                    future.set_exception(error)



from typing import (
    Any,
    Callable,
    Optional,
    List,
)

from .exceptions import InvalidStateError

_PENDING = "PENDING"
_FINISHED = "FINISHED"


class SyncFuture:

    _state = _PENDING
    _result: Optional[Any] = None
    _exception: Optional[Exception] = None
    _callbacks: List[Callable]
    _cancel_message = None

    deferred_callback: Optional[Callable] = None

    def __init__(self):
        self._callbacks = []

    def done(self) -> bool:
        return self._state != _PENDING

    def result(self):
        self._assert_state(_FINISHED)
        if self._exception is not None:
            raise self._exception
        return self._result

    def exception(self):
        self._assert_state(_FINISHED)
        return self._exception

    def add_done_callback(self, fn: Callable) -> None:
        self._assert_state(_PENDING)
        self._callbacks.append(fn)

    def set_result(self, result: Any) -> None:
        self._assert_state(_PENDING)
        self._result = result
        self._finish()

    def set_exception(self, exception: Exception) -> None:
        self._assert_state(_PENDING)
        if isinstance(exception, type):
            exception = exception()
        self._exception = exception
        self._finish()

    def _assert_state(self, state: str) -> None:
        if self._state != state:
            raise InvalidStateError(f"Future is not {state}")

    def _finish(self):
        self._state = _FINISHED
        callbacks = self._callbacks
        if not callbacks:
            return
        self._callbacks = []
        for callback in callbacks:
            callback()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graphql_sync_dataloaders']

package_data = \
{'': ['*']}

install_requires = \
['graphql-core>=3.2.0,<3.3.0']

setup_kwargs = {
    'name': 'graphql-sync-dataloaders',
    'version': '0.1.0',
    'description': 'Use DataLoaders in your Python GraphQL servers that have to run in a sync context (i.e. Django).',
    'long_description': '# graphql-sync-dataloaders\n\nUse DataLoaders in your Python GraphQL servers that have to run in a sync\ncontext (i.e. Django).\n\n## Requirements\n\n* Python 3.8+\n* graphql-core >=3.2.0\n\n## Installation\n\nThis package can be installed from [PyPi](https://pypi.python.org/pypi/graphql-sync-dataloaders) by running:\n\n```bash\npip install graphql-sync-dataloaders\n```\n\n## Strawberry setup\n\nWhen creating your Strawberry Schema pass `DeferredExecutionContext` as the\n`execution_context_class` argument:\n\n```python\n# schema.py\nimport strawberry\nfrom graphql_sync_dataloaders import DeferredExecutionContext\n\nschema = strawberry.Schema(Query, execution_context_class=DeferredExecutionContext)\n```\n\nThen create your dataloaders using the `SyncDataLoader` class:\n\n```python\nfrom typing import List\n\nfrom graphql_sync_dataloaders import SyncDataLoader\n\nfrom .app import models  # your Django models\n\ndef load_users(keys: List[int]) -> List[User]:\n    qs = models.User.objects.filter(id__in=keys)\n    user_map = {user.id: user for user in qs}\n    return [user_map.get(key, None) for key in keys]\n\nuser_loader = SyncDataLoader(load_users)\n```\n\nYou can then use the loader in your resolvers and it will automatically be\nbatched to reduce the number of SQL queries:\n\n```python\nimport strawberry\n\n@strawberry.type\nclass Query:\n    @strawberry.field\n    def get_user(self, id: strawberry.ID) -> User:\n        return user_loader.load(id)\n```\n\n**Note: You probably want to setup your loaders in context. See\nhttps://strawberry.rocks/docs/guides/dataloaders#usage-with-context for more\ndetails**\n\nThe following query will only make 1 SQL query:\n\n```graphql\nfragment UserDetails on User {\n  username\n}\n\nquery {\n  user1: getUser(id: \'1\') {\n    ...UserDetails\n  }\n  user2: getUser(id: \'2\') {\n    ...UserDetails\n  }\n  user3: getUser(id: \'3\') {\n    ...UserDetails\n  }\n}\n```\n\n\n## Graphene-Django setup\n\n**Requires graphene-django >=3.0.0b8**\n\nWhen setting up your GraphQLView pass `DeferredExecutionContext` as the\n`execution_context_class` argument:\n\n```python\n# urls.py\nfrom django.urls import path\nfrom graphene_django.views import GraphQLView\nfrom graphql_sync_dataloaders import DeferredExecutionContext\n\nfrom .schema import schema\n\nurlpatterns = [\n    path(\n        "graphql",\n        csrf_exempt(\n            GraphQLView.as_view(\n                schema=schema, \n                execution_context_class=DeferredExecutionContext\n            )\n        ),\n    ),\n]\n```\n\nThen create your dataloaders using the `SyncDataLoader` class:\n\n```python\nfrom typing import List\n\nfrom graphql_sync_dataloaders import SyncDataLoader\n\nfrom .app import models  # your Django models\n\ndef load_users(keys: List[int]) -> List[User]:\n    qs = models.User.objects.filter(id__in=keys)\n    user_map = {user.id: user for user in qs}\n    return [user_map.get(key, None) for key in keys]\n\nuser_loader = SyncDataLoader(load_users)\n```\n\nYou can then use the loader in your resolvers and it will automatically be\nbatched to reduce the number of SQL queries:\n\n```python\nimport graphene\n\nclass Query(graphene.ObjectType):\n    get_user = graphene.Field(User, id=graphene.ID)\n\n    def resolve_get_user(root, info, id):\n        return user_loader.load(id)\n```\n\nThe following query will only make 1 SQL query:\n\n```graphql\nfragment UserDetails on User {\n  username\n}\n\nquery {\n  user1: getUser(id: \'1\') {\n    ...UserDetails\n  }\n  user2: getUser(id: \'2\') {\n    ...UserDetails\n  }\n  user3: getUser(id: \'3\') {\n    ...UserDetails\n  }\n}\n```\n\n## How it works\n\nThis library implements a custom version of the graphql-core\n[ExecutionContext class](https://github.com/graphql-python/graphql-core/blob/5f6a1944cf6923f6249d1575f5b3aad87e629c66/src/graphql/execution/execute.py#L171)\nthat is aware of the `SyncFuture` objects defined in this library. A\n`SyncFuture` represents a value that hasn\'t been resolved to a value yet\n(similiar to asycnio Futures or JavaScript Promises) and that is what the\n`SyncDataLoader` returns when you call the `.load` function.\n\nWhen the custom `ExecutionContext` encounters a `SyncFuture` that gets returned\nfrom a resolver and it keeps track of them. Then after the first pass of the\nexection it triggers the `SyncFuture` callbacks until there are none left. Once\nthere are none left the data is fully resolved and can be returned to the\ncaller synchronously. This allows us to implement a `DataLoader` pattern that\nbatches calls to a loader function, and it allows us to do this in a fully\nsynchronously way.\n\n## Credits\n\n[@Cito](https://github.com/Cito) for graphql-core and for implementing the first version of this in https://github.com/graphql-python/graphql-core/pull/155\n',
    'author': 'Jonathan Kim',
    'author_email': 'hello@jkimbo.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import (
    Any,
    AsyncIterable,
    Callable,
    Dict,
    Optional,
    List,
    Iterable,
    Union,
    cast,
)
from functools import partial

from graphql import (
    ExecutionContext,
    FieldNode,
    GraphQLError,
    GraphQLObjectType,
    GraphQLOutputType,
    GraphQLResolveInfo,
    GraphQLList,
    OperationDefinitionNode,
    located_error,
)
from graphql.pyutils import (
    is_iterable,
    Path,
    AwaitableOrValue,
    Undefined,
)

from graphql.execution.execute import get_field_def
from graphql.execution.values import get_argument_values

from .sync_future import SyncFuture


class DeferredExecutionContext(ExecutionContext):
    """Execution for working with synchronous Futures.

    This execution context can handle synchronous Futures as resolved values.
    Deferred callbacks set in these Futures are called after the operation
    is executed and before the result is returned.
    """

    _deferred_callbacks: List[Callable]

    def execute_operation(
        self, operation: OperationDefinitionNode, root_value: Any
    ) -> Optional[AwaitableOrValue[Any]]:
        self._deferred_callbacks = []
        result = super().execute_operation(operation, root_value)

        callbacks = self._deferred_callbacks
        while callbacks:
            callbacks.pop(0)()

        if isinstance(result, SyncFuture):
            if not result.done():
                raise RuntimeError("GraphQL deferred execution failed to complete.")
            return result.result()

        return result

    def execute_fields_serially(
        self,
        parent_type: GraphQLObjectType,
        source_value: Any,
        path: Optional[Path],
        fields: Dict[str, List[FieldNode]],
    ) -> Union[AwaitableOrValue[Dict[str, Any]], SyncFuture]:
        results: AwaitableOrValue[Dict[str, Any]] = {}

        unresolved = 0
        for response_name, field_nodes in fields.items():
            field_path = Path(path, response_name, parent_type.name)
            result = self.execute_field(
                parent_type, source_value, field_nodes, field_path
            )
            if isinstance(result, SyncFuture):
                if result.done():
                    result = result.result()
                    if result is not Undefined:
                        results[response_name] = result
                else:

                    # noinspection PyShadowingNames, PyBroadException
                    def process_result(response_name: str, result: SyncFuture) -> None:
                        nonlocal unresolved
                        awaited_result = result.result()
                        if awaited_result is not Undefined:
                            results[response_name] = awaited_result
                        unresolved -= 1
                        if not unresolved:
                            future.set_result(results)

                    unresolved += 1
                    result.add_done_callback(
                        partial(process_result, response_name, result)
                    )
            elif result is not Undefined:
                results[response_name] = result

        if not unresolved:
            return results

        future = SyncFuture()
        return future

    execute_fields = execute_fields_serially

    def execute_field(
        self,
        parent_type: GraphQLObjectType,
        source: Any,
        field_nodes: List[FieldNode],
        path: Path,
    ) -> AwaitableOrValue[Any]:
        field_def = get_field_def(self.schema, parent_type, field_nodes[0])
        if not field_def:
            return Undefined
        return_type = field_def.type
        resolve_fn = field_def.resolve or self.field_resolver
        if self.middleware_manager:
            resolve_fn = self.middleware_manager.get_field_resolver(resolve_fn)
        info = self.build_resolve_info(field_def, field_nodes, parent_type, path)
        try:
            args = get_argument_values(field_def, field_nodes[0], self.variable_values)
            result = resolve_fn(source, info, **args)

            if isinstance(result, SyncFuture):

                if result.done():
                    completed = self.complete_value(
                        return_type, field_nodes, info, path, result.result()
                    )

                else:

                    callback = result.deferred_callback
                    if callback:
                        self._deferred_callbacks.append(callback)

                    # noinspection PyShadowingNames
                    def process_result():
                        try:
                            completed = self.complete_value(
                                return_type, field_nodes, info, path, result.result()
                            )
                            if isinstance(completed, SyncFuture):

                                # noinspection PyShadowingNames
                                def process_completed():
                                    try:
                                        future.set_result(completed.result())
                                    except Exception as raw_error:
                                        error = located_error(
                                            raw_error, field_nodes, path.as_list()
                                        )
                                        self.handle_field_error(error, return_type)
                                        future.set_result(None)

                                if completed.done():
                                    process_completed()
                                else:
                                    completed.add_done_callback(process_completed)
                            else:
                                future.set_result(completed)
                        except Exception as raw_error:
                            error = located_error(
                                raw_error, field_nodes, path.as_list()
                            )
                            self.handle_field_error(error, return_type)
                            future.set_result(None)

                    future = SyncFuture()
                    result.add_done_callback(process_result)
                    return future

            else:
                completed = self.complete_value(
                    return_type, field_nodes, info, path, result
                )

            if isinstance(completed, SyncFuture):

                # noinspection PyShadowingNames
                def process_completed():
                    try:
                        future.set_result(completed.result())
                    except Exception as raw_error:
                        error = located_error(raw_error, field_nodes, path.as_list())
                        self.handle_field_error(error, return_type)
                        future.set_result(None)

                if completed.done():
                    return process_completed()

                future = SyncFuture()
                completed.add_done_callback(process_completed)
                return future

            return completed
        except Exception as raw_error:
            error = located_error(raw_error, field_nodes, path.as_list())
            self.handle_field_error(error, return_type)
            return None

    def complete_list_value(
        self,
        return_type: GraphQLList[GraphQLOutputType],
        field_nodes: List[FieldNode],
        info: GraphQLResolveInfo,
        path: Path,
        result: Union[AsyncIterable[Any], Iterable[Any]],
    ) -> Union[AwaitableOrValue[List[Any]], SyncFuture]:
        if not is_iterable(result):
            if isinstance(result, SyncFuture):

                def process_result():
                    return self.complete_list_value(
                        return_type, field_nodes, info, path, result.result()
                    )

                if result.done():
                    return process_result()
                future = SyncFuture()
                result.add_done_callback(process_result)
                return future

            raise GraphQLError(
                "Expected Iterable, but did not find one for field"
                f" '{info.parent_type.name}.{info.field_name}'."
            )
        result = cast(Iterable[Any], result)

        item_type = return_type.of_type
        results: List[Any] = [None] * len(result)

        unresolved = 0

        for index, item in enumerate(result):
            item_path = path.add_key(index, None)

            try:
                if isinstance(item, SyncFuture):

                    if item.done():
                        completed = self.complete_value(
                            item_type, field_nodes, info, item_path, item.result()
                        )
                    else:
                        callback = item.deferred_callback
                        if callback:
                            self._deferred_callbacks.append(callback)

                        # noinspection PyShadowingNames
                        def process_item(
                            index: int, item: SyncFuture, item_path: Path
                        ) -> None:
                            nonlocal unresolved
                            try:
                                completed = self.complete_value(
                                    item_type,
                                    field_nodes,
                                    info,
                                    item_path,
                                    item.result(),
                                )
                                if isinstance(completed, SyncFuture):
                                    if completed.done():
                                        results[index] = completed.result()
                                    else:

                                        # noinspection PyShadowingNames
                                        def process_completed(
                                            index: int,
                                            completed: SyncFuture,
                                            item_path: Path,
                                        ) -> None:
                                            try:
                                                results[index] = completed.result()
                                            except Exception as raw_error:
                                                error = located_error(
                                                    raw_error,
                                                    field_nodes,
                                                    item_path.as_list(),
                                                )
                                                self.handle_field_error(
                                                    error, item_type
                                                )

                                        completed.add_done_callback(
                                            partial(
                                                process_completed,
                                                index,
                                                completed,
                                                item_path,
                                            )
                                        )
                                else:
                                    results[index] = completed
                            except Exception as raw_error:
                                error = located_error(
                                    raw_error, field_nodes, item_path.as_list()
                                )
                                self.handle_field_error(error, item_type)
                            unresolved -= 1
                            if not unresolved:
                                future.set_result(results)

                        unresolved += 1
                        item.add_done_callback(
                            partial(process_item, index, item, item_path)
                        )
                        continue
                else:
                    completed = self.complete_value(
                        item_type, field_nodes, info, item_path, item
                    )

                if isinstance(completed, SyncFuture):

                    if completed.done():
                        results[index] = completed.result()
                    else:
                        callback = completed.deferred_callback
                        if callback:
                            self._deferred_callbacks.append(callback)

                        # noinspection PyShadowingNames
                        def process_completed(
                            index: int, completed: SyncFuture, item_path: Path
                        ) -> None:
                            nonlocal unresolved
                            try:
                                results[index] = completed.result()
                            except Exception as raw_error:
                                error = located_error(
                                    raw_error, field_nodes, item_path.as_list()
                                )
                                self.handle_field_error(error, item_type)
                            unresolved -= 1
                            if not unresolved:
                                future.set_result(results)

                        unresolved += 1
                        completed.add_done_callback(
                            partial(process_completed, index, completed, item_path)
                        )
                else:
                    results[index] = completed
            except Exception as raw_error:
                error = located_error(raw_error, field_nodes, item_path.as_list())
                self.handle_field_error(error, item_type)

        if not unresolved:
            return results

        future = SyncFuture()
        return future



from .execution_context import DeferredExecutionContext
from .sync_dataloader import SyncDataLoader
from .sync_future import SyncFuture, InvalidStateError


__all__ = [
    "DeferredExecutionContext",
    "SyncDataLoader",
    "SyncFuture",
    "InvalidStateError",
]
